/*
 * Ind_CPPN.hpp
 *
 *  Created on: Mar 17, 2015
 *      Author: Joost Huizinga
 */

#ifndef IND_CPPN_HPP_
#define IND_CPPN_HPP_

#include <vector>
#include <set>
#include <list>
#include <cmath>
#include <fstream>

#include "Misc_Random.hpp"
#include "Ind_CPPN_Neuron.hpp"
#include "Ind_NeuralNetwork.hpp"
#include "Misc_BreadthFirstSearch.hpp"


/**
 * A neural network designed to evolve Compositional Pattern Producing Networks (CPPNs)
 *
 * The constructor and mutation functions are inspired by NEAT complexification,
 * such that networks start with a fully connected set of inputs and outputs,
 * but without hidden neurons. Additional neurons can be added later through the mutation
 * function.
 *
 * This network does not implement crossover, and it does not support speciation based
 * on historical markings.
 */
class CPPN: public NeuralNetwork<CPPN_Neuron, Connection>
{
public:

    /**
     * Initializes the network with the supplied number of inputs and outputs.
     *
     * Extend this function such that the network starts with the supplied number
     * of inputs and outputs, and such that each input is connected to every output.
     *
     * In addition, this function should set a default value for the _neuronActMutRate
     * mutation rate (0.05 is suggested).
     *
     * Remember that most other values should already be set the the constructor of
     * your neural network, which is called by this line of code:
     *   NeuralNetwork<CPPN_Neuron, Connection>(numberOfInputs + numberOfOutputs)
     */
    CPPN(size_t numberOfInputs = 0, size_t numberOfOutputs = 0):
        NeuralNetwork<CPPN_Neuron, Connection>(numberOfInputs + numberOfOutputs)
    {
        //TODO: YOUR CODE HERE

    }


    /*******************************************
     ************* Debug functions *************
     *******************************************/

    /**
     * Simple debug print function to print all neurons and connections in the network.
     */
    void debugPrint(){
        //Write neurons
        std::cout << "Neurons: " << std::endl;
        for(size_t i=0; i<_neurons.size(); ++i){
            std::cout << "Bias: "<< _neurons[i].getBias() << " ";
            std::cout << " activation function: ";
            switch(_neurons[i].getActivationFunction()){
            case CPPN_Neuron::linear:
                std::cout << "lin";
                break;
            case CPPN_Neuron::sine:
                std::cout << "sin";
                break;
            case CPPN_Neuron::gaussian:
                std::cout << "gau";
                break;
            case CPPN_Neuron::sigmoid:
                std::cout << "sig";
                break;
            default:
                std::cerr << "Unknown activation function" << std::endl;
            }
            std::cout << std::endl;
        }

        //Write connections
        std::cout << "Connections: " << std::endl;
        for(size_t i=0; i<_connections.size(); ++i){
            std::cout << "Source :" << _connections[i].getSource()
                    << " target: " << _connections[i].getTarget()
                    << " weight: "<< _connections[i].getWeight() << std::endl;
        }
    }

    /**
     * Assigns random weights to each connection and random biases and activation functions to each neuron.
     *
     * Implement this function such that each connection is assigned, uniform randomly,
     * a number in [_minWeight, _maxWeight], and such that each neuron is assigned,
     * uniform randomly, a bias in [_minWeight, _maxWeight], and, uniform randomly,
     * one of the four activation functions (linear, sin, gaussian, sigmoid).
     *
     * Do not assign biases or activation functions to the input neurons, these neurons
     * are set by sensors and thus these values should be ignored.
     */
    void randomize(){
        //TODO: YOUR CODE HERE

    }


    /**
     * Mutates this network.
     *
     * The same as the regular neural network implementation, with the following exceptions:
     * - Neurons have an activation function that needs to be mutated. For each neuron, with
     *   a probability of _neuronActMutRate, replace the current activation function with a different
     *   activation function, uniform randomly selected from (linear, sin, gaussian, sigmoid).
     *
     * - The CPPN is a feed-forward network, meaning that recurrent connections are not allowed.
     *   You have to make sure that, when you randomly add a connection, this mutation does not cause a
     *   cycle in the network.
     *   To do so you can start by selecting a source node at random. Now create a set of nodes that:
     *   - does not include any of the inputs.
     *   - does not include any ancestors of the source node
     *   - does not include any nodes already directly connected to the source node
     *   You can now randomly select a node from this set, and be guaranteed that the network does not contain
     *   cycles. If the set turns out to be empty you can ignore this mutation event.
     *
     * You can use the BreadthFirstItereator to help you, but do keep in mind
     * that this iterator will get stuck in an infinite loop if your network contains cycles.
     */
    void mutate(){
        //TODO: YOUR CODE HERE

    }

    /**
     * Calculates the 'depth' of this network, where the depth is the longest path
     * from any input to any output.
     *
     * Implement such that it returns the length of the longest path between any input and
     * any output.
     *
     * You can use the BreadthFirstItereator to help you, but do keep in mind
     * that this iterator will get stuck in an infinite loop if your network contains cycles.
     */
    size_t getDepth(){
        //TODO: YOUR CODE HERE

    }

protected:
    double _neuronActMutRate;

    //TODO: MORE CODE HERE IF YOU WANT

};

/**
 * Convenience function for writing network connections to a file-stream.
 */
std::ostream& operator<<(std::ostream& is, CPPN& obj){
    std::vector<CPPN_Neuron> neurons = obj.getNeurons();
    std::vector<Connection> connections = obj.getConnections();

    //Write neurons
    is << neurons.size() << " ";
    is << obj.getNbOfInputs() << " ";
    is << obj.getNbOfOutputs() << " ";
    is << connections.size() << " ";
    for(size_t i=0; i<neurons.size(); ++i){
        is << neurons[i].getBias() << " ";
    }

    //Write connections
    for(size_t i=0; i<connections.size(); ++i){
        is << connections[i].getSource() << " " << connections[i].getTarget() << " "<< connections[i].getWeight()<< " ";
    }
    return is;
}




#endif /* IND_CPPN_HPP_ */
